/***************************************************************************
 *   Copyright (C) 2004 by Fabian Vilers                                   *
 *   fabian@vilers.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qstringlist.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qlistview.h>
#include <qapplication.h>
#include <qclipboard.h>
#include <qprogressbar.h>
#include <qregexp.h>
#include <qgroupbox.h>
#include <qradiobutton.h>
#include <qsettings.h>
#include <qstatusbar.h>

#include "passwordcreator.h"
#include "aboutbox.h"

/*
 *  Public functions
 */
PasswordCreator::PasswordCreator(QWidget *parent, const char *name, WFlags fl)
        : PasswordCreatorForm(parent, name, fl)
{
    // Seed the random number generator with time
    srand(time(NULL) + getpid());
    
    // Hide status bar
    statusBar()->hide();
    
    // Restore settings
    QSettings settings;
    QString keyPrefix = "passwordcreator/";
    
    // Window
    setMinimumHeight(settings.readNumEntry(keyPrefix + "window/height", 464));
    setMinimumWidth(settings.readNumEntry(keyPrefix + "window/width", 407));
    
    if (settings.readNumEntry(keyPrefix + "settings/numbers", 0) > 0) numbersCheck->setChecked(true);
    if (settings.readNumEntry(keyPrefix + "settings/lowercase", 0) > 0) lowercaseCheck->setChecked(true);
    if (settings.readNumEntry(keyPrefix + "settings/uppercase", 0) > 0) uppercaseCheck->setChecked(true);
    if (settings.readNumEntry(keyPrefix + "settings/signs", 0) > 0) signsCheck->setChecked(true);
    
    if (settings.readNumEntry(keyPrefix + "settings/specific", 0) > 0)
    {
        specificCheck->setChecked(true);
        specificEdit->setText(settings.readEntry(keyPrefix + "settings/specific_text"));
    }
    
    if (settings.readNumEntry(keyPrefix + "settings/file", 0) > 0)
    {
        // Set the text before the check to avoid the file dialog to open
        fileEdit->setText(settings.readEntry(keyPrefix + "settings/file_text"));
        fileCheck->setChecked(true);
    }
    
    if (settings.readNumEntry(keyPrefix + "settings/exclude", 0) > 0)
    {
        specificExcludeCheck->setChecked(true);
        specificExcludeEdit->setText(settings.readEntry(keyPrefix + "settings/exclude_text"));
    }
    
    lengthSpin->setValue(settings.readNumEntry(keyPrefix + "settings/length", 8));
    quantitySpin->setValue(settings.readNumEntry(keyPrefix + "settings/quantity", 1));
}

PasswordCreator::~PasswordCreator()
{
}

void PasswordCreator::generateButton_clicked()
{
    QString fileContent = "";
    QString password = "";
        
    // Read the file content if it was checked
    if (!fileEdit->text().isEmpty())
    {
        QFile file(fileEdit->text());
        if (file.open(IO_ReadOnly))
        {
            QTextStream stream(&file);
            fileContent = stream.read();
        }
    }
    
    // Generate the number of password asked
    for (int i = 0; i < quantitySpin->value(); i++)
    {
        if (customRadio->isChecked())
        {
            password = makePassword(lengthSpin->value(),
                                    numbersCheck->isChecked(),
                                    lowercaseCheck->isChecked(),
                                    uppercaseCheck->isChecked(),
                                    signsCheck->isChecked(),
                                    specificExcludeEdit->text(),
                                    specificEdit->text(),
                                    fileContent);
        }
        else
        {
            password = makePronouncable(lengthSpin->value(), specificExcludeEdit->text());
        }
        
        // Maybe all the option were to false so we must be sure to have returned a non blank password
        if (!password.isEmpty())
        {
            int quality = evaluateQuality(password);
            QString evaluation = "";
            
            if (quality < 1) evaluation = "Weak";
            else if (quality < 2) evaluation = "Bad";
            else if (quality < 3) evaluation = "Good";
            else if (quality < 4) evaluation = "Strong";
            else evaluation = "Excellent";
            
            passwordList->insertItem(new QListViewItem(passwordList, password, evaluation));
        }
    }
}

void PasswordCreator::fileCheck_toggled(bool state)
{
    fileEdit->setEnabled(state);
    
    if (state)
    {
        // Show the file dialog immediately after the check click is done
        if (fileEdit->text().isEmpty()) fileEdit->setText(QFileDialog::getOpenFileName("", "Text (*.asc *.txt *.text)", this, "", "Choose a file containing specific characters"));
        
        // Clear the check if no file was specified
        if (fileEdit->text().isEmpty()) fileCheck->setChecked(false);
    }
    else
    {
        // No need to keep the old value
        fileEdit->clear();
    }
}

void PasswordCreator::specificCheck_toggled(bool state)
{
    specificEdit->setEnabled(state);
    
    // No need to keep the old value
    specificEdit->clear();
}

void PasswordCreator::clearButton_clicked()
{
    // Clear the list
    passwordList->clear();
}

void PasswordCreator::clipboardButton_clicked()
{
    QString selected = "";
    
    // Find wich item are selected and append them in a string
    QListViewItemIterator it(passwordList);
    while (it.current())
    {
        // Copy text from the password (0) column
        if (it.current()->isSelected()) selected += it.current()->text(0) + '\n';
        ++it;
    }
    
    // If we have selected item, paste the string to clipboard and remove trailing newline
    if (!selected.isEmpty()) QApplication::clipboard()->setText(selected.left(selected.length() - 1));
}

void PasswordCreator::customRadio_toggled(bool state)
{
    includeBox->setEnabled(state);
}

void PasswordCreator::fileExit()
{
    close();
}

void PasswordCreator::helpAbout()
{
    AboutBox(VERSION).exec();
}

void PasswordCreator::specificExcludeCheck_toggled(bool state)
{
    specificExcludeEdit->setEnabled(state);
    
    // No need to keep the old value
    specificExcludeEdit->clear();
}

/*
 *  Protected slots
 */
void PasswordCreator::closeEvent(QCloseEvent *e)
{
    // Save settings
    QSettings settings;
    QString keyPrefix = "passwordcreator/";
    
    // Checkbox
    if (numbersCheck->isChecked()) settings.writeEntry(keyPrefix + "settings/numbers", 1);
    else settings.writeEntry(keyPrefix + "settings/numbers", 0);
    
    if (lowercaseCheck->isChecked()) settings.writeEntry(keyPrefix + "settings/lowercase", 1);
    else settings.writeEntry(keyPrefix + "settings/lowercase", 0);
    
    if (uppercaseCheck->isChecked()) settings.writeEntry(keyPrefix + "settings/uppercase", 1);
    else settings.writeEntry(keyPrefix + "settings/uppercase", 0);
    
    if (signsCheck->isChecked()) settings.writeEntry(keyPrefix + "settings/signs", 1);
    else settings.writeEntry(keyPrefix + "settings/signs", 0);
    
    if (specificCheck->isChecked()) settings.writeEntry(keyPrefix + "settings/specific", 1);
    else settings.writeEntry(keyPrefix + "settings/specific", 0);
    settings.writeEntry(keyPrefix + "settings/specific_text", specificEdit->text());
    
    if (fileCheck->isChecked()) settings.writeEntry(keyPrefix + "settings/file", 1);
    else settings.writeEntry(keyPrefix + "settings/file", 0);
    settings.writeEntry(keyPrefix + "settings/file_text", fileEdit->text());
    
    if (specificExcludeCheck->isChecked()) settings.writeEntry(keyPrefix + "settings/exclude", 1);
    else settings.writeEntry(keyPrefix + "settings/exclude", 0);
    settings.writeEntry(keyPrefix + "settings/exclude_text", specificExcludeEdit->text());
    
    // Length and quantity
    settings.writeEntry(keyPrefix + "settings/length", lengthSpin->value());
    settings.writeEntry(keyPrefix + "settings/quantity", quantitySpin->value());
    
    // Windows geometry
    settings.writeEntry(keyPrefix + "window/height", height());
    settings.writeEntry(keyPrefix + "window/width", width());
    
    // Close
    return(QMainWindow::closeEvent(e));
}

/*
 *  Private functions
 */
QString PasswordCreator::makePassword(int length, bool withNumbers, bool withLowercase, bool withUppercase, bool withSigns, QString exclude, QString specific, QString file)
{
    QStringList pool;
    QString password = "";
    
    // Create characters pool
    if (withNumbers) pool.append(removeExcluded(numbers, exclude));
    if (withLowercase) pool.append(removeExcluded(characters, exclude));
    if (withUppercase) pool.append(removeExcluded(characters.upper(), exclude));
    if (withSigns) pool.append(removeExcluded(signs, exclude));
    if (!specific.isEmpty()) pool.append(specific);
    if (!file.isEmpty()) pool.append(file);
    
    if (pool.count() > 0)
    {
        for (int i = 0; i < length; i++)
        {
            // Random wich pool to use
            int poolType = rand() % pool.count();
            
            // Select a character
            int charPos = rand() % pool[poolType].length();
            
            // Add it to password
            password += pool[poolType][charPos];
        }
    }
    
    return(password);
}

int PasswordCreator::evaluateQuality(QString password)
{
    int minimum = ((password.length() + 3) / 4);
    QRegExp numbersRegexp("[" + numbers + "]", TRUE, FALSE);
    QRegExp lowercaseRegexp("[" + characters + "]", TRUE, FALSE);
    QRegExp uppercaseRegexp("[" + characters.upper() + "]", TRUE, FALSE);
    QRegExp signsRegexp("[" + protect(signs) + "]", TRUE, FALSE);
    int quality = 0;
    
    if (password.contains(numbersRegexp) >= minimum) quality++;
    if (password.contains(lowercaseRegexp) >= minimum) quality++;
    if (password.contains(uppercaseRegexp) >= minimum) quality++;
    if (password.contains(signsRegexp) >= minimum) quality++;
    if ((password.length() >= 12) && (quality >= 2)) quality++;
    
    return(quality);
}

QString PasswordCreator::makePronouncable(int length, QString exclude)
{
    QString password = "";
    int nbConsonants = 0;
    int nbVowels = 0;
    QString vowelsPool = vowels;
    QString consonantsPool = consonants;
    
    // Exclude some characters
    exclude += "yxz";
    for (unsigned int i = 0; i < exclude.length(); i++)
    {
        vowelsPool.remove(exclude[i]);
        consonantsPool.remove(exclude[i]);
    }
    
    // TODO must improve
    if ((vowelsPool.length() > 0) && (consonantsPool.length() > 0))
    {
        for (int i = 0; i < (length - 1); i++)
        {
            // Select wich letter to use
            int letterType = 0;
            if ((i == 1) && (nbVowels == 1)) letterType = CONSONANT;
            else if (nbConsonants == 2) letterType = VOWEL;
            else if (nbVowels == 2) letterType = CONSONANT;
            else letterType = rand() % 2;
            
            // Select a character
            if (letterType == VOWEL)
            {
                int charPos = rand() % vowelsPool.length();
                password += vowelsPool[charPos];
                nbConsonants = 0;
                nbVowels++;
            }
            else
            {
                int charPos = rand() % consonantsPool.length();
                password += consonantsPool[charPos];
                nbVowels = 0;
                nbConsonants++;
            }
        }
        
        // Do not finish with two consonants
        if (consonants.contains(password.right(1)) > 0)
        {
            int charPos = rand() % vowelsPool.length();
            password += vowelsPool[charPos];
        }
        else
        {
            int charPos = rand() % consonantsPool.length();
            password += consonantsPool[charPos];
        }
    }
    
    return(password);
}

QString PasswordCreator::removeExcluded(QString source, QString exclude)
{
    for (unsigned int i = 0; i < exclude.length(); i++)
    {
        source.remove(exclude[i]);
    }
    
    return(source);
}

QString PasswordCreator::protect(QString source)
{
    QString ret = "";
    
    for (unsigned int i = 0; i < source.length(); i++)
    {
        ret += "\\" + source[i];
    }
    
    return(ret);
}
